"""Connection"""

import logging as _logging
import time as _time


class Connection():
    """Connection"""

    DEFAULT_SEND_TIMEOUT = 5.0
    DEFAULT_BUFFER_LIMIT = None

    def __init__(
            self, connection, send_timeout=None, buffer_limit=None,
            log=None):
        self._log = log if log else _logging.Logger(self.__class__.__name__)
        self._socket, self._addr = connection
        self._out_buffer = bytearray()
        self._last_write_time = _time.time()
        if send_timeout is not None:
            self._send_timeout = send_timeout
        else:
            self._send_timeout = self.DEFAULT_SEND_TIMEOUT
        if buffer_limit is not None:
            self._buffer_limit = buffer_limit
        else:
            self._buffer_limit = self.DEFAULT_BUFFER_LIMIT

    def __del__(self):
        self.close()

    def socket(self):
        """Return reference to socket"""
        return self._socket

    def close(self):
        """Close connection"""
        if self._socket:
            self._socket.close()
            self._socket = None
            self._log.info("Client disconnected: %s:%d", *self._addr)

    def fileno(self):
        """emulate fileno method of socket"""
        return self._socket.fileno() if self._socket else None

    def get_address(self):
        """Return address"""
        return self._addr

    def send(self, data):
        """Add data to output buffer, return number of bytes added"""
        if not self._socket:
            return None
        new_size = len(self._out_buffer) + len(data)
        if self._buffer_limit and new_size > self._buffer_limit:
            return None
        if not self._out_buffer:
            # Reset timeout when buffer becomes non-empty
            self._last_write_time = _time.time()
        self._out_buffer.extend(data)
        return len(data)

    def flush(self):
        """Flush output buffer, return number of bytes sent or None on error"""
        if not self._socket or not self._out_buffer:
            return 0
        try:
            sent = self._socket.send(self._out_buffer)
            if sent > 0:
                del self._out_buffer[:sent]
                self._last_write_time = _time.time()
            return sent
        except OSError:
            return None

    def has_pending_data(self):
        """Return True if there is data in output buffer"""
        return bool(self._out_buffer)

    def is_stale(self):
        """Return True if send timeout expired"""
        if not self._out_buffer:
            return False
        return _time.time() - self._last_write_time > self._send_timeout



"""Server"""

# pylint: disable=C0209

import socket as _socket
import logging as _logging
import ser2tcp.connection_tcp as _connection_tcp
import ser2tcp.connection_telnet as _connection_telnet


class ConfigError(Exception):
    """Configuration error exception"""


class Server():
    """Server connection manager"""

    CONNECTIONS = {
        'TCP': _connection_tcp.ConnectionTcp,
        'TELNET': _connection_telnet.ConnectionTelnet,
    }

    def __init__(self, config, ser, log=None):
        self._log = log if log else _logging.Logger(self.__class__.__name__)
        self._config = config
        self._serial = ser
        self._connections = []
        self._protocol = self._config['protocol'].upper()
        self._send_timeout = self._config.get('send_timeout')
        self._buffer_limit = self._config.get('buffer_limit')
        self._socket = None
        self._log.info(
            "  Server: %s %d %s",
            self._config['address'],
            self._config['port'],
            self._protocol)
        if self._protocol not in self.CONNECTIONS:
            raise ConfigError('Unknown protocol %s' % self._protocol)
        self._socket = _socket.socket(
            _socket.AF_INET, _socket.SOCK_STREAM, _socket.IPPROTO_TCP)
        self._socket.setsockopt(_socket.SOL_SOCKET, _socket.SO_REUSEADDR, 1)
        self._socket.bind((config['address'], config['port']))
        self._socket.listen(1)

    def __del__(self):
        self.close()

    def _client_connect(self):
        """connect to client, will accept waiting connection"""
        sock, addr = self._socket.accept()
        if not self._connections:
            if not self._serial.connect():
                self._log.info("Client canceled: %s:%d", *addr)
                sock.close()
                return
        connection = self.CONNECTIONS[self._protocol](
            connection=(sock, addr),
            ser=self._serial,
            send_timeout=self._send_timeout,
            buffer_limit=self._buffer_limit,
            log=self._log,
        )
        if self._serial.connect():
            self._connections.append(connection)
        else:
            connection.close()

    def close_connections(self):
        """close all clients"""
        while self._connections:
            self._connections.pop().close()

    def close(self):
        """Close socket and all connections"""
        if self._socket is not None:
            self.close_connections()
            self._socket.close()
            self._socket = None

    def has_connections(self):
        """True if server has some connections"""
        return bool(self._connections)

    def read_sockets(self):
        """Return sockets for reading (server + all clients)"""
        sockets = [self._socket]
        for con in self._connections:
            sockets.append(con.socket())
        return sockets

    def write_sockets(self):
        """Return sockets for writing (clients with pending data)"""
        sockets = []
        for con in self._connections:
            if con.has_pending_data():
                sockets.append(con.socket())
        return sockets

    def _remove_connection(self, con):
        """Remove connection and disconnect serial if no connections left"""
        con.close()
        self._connections.remove(con)
        if not self._connections:
            self._serial.disconnect()

    def process_read(self, read_sockets):
        """Process sockets with read event"""
        if self._socket in read_sockets:
            self._client_connect()
        for con in list(self._connections):
            if con.socket() in read_sockets:
                data = b''
                try:
                    data = con.socket().recv(4096)
                    self._log.debug("(%s:%d): %s", *con.get_address(), data)
                except ConnectionResetError as err:
                    self._log.info("(%s:%d): %s", *con.get_address(), err)
                if not data:
                    self._remove_connection(con)
                    continue
                con.on_received(data)

    def process_write(self, write_sockets):
        """Process sockets with write event, flush buffers"""
        for con in list(self._connections):
            if con.socket() in write_sockets:
                result = con.flush()
                if result is None:
                    self._log.info(
                        "(%s:%d): write error", *con.get_address())
                    self._remove_connection(con)

    def process_stale(self):
        """Remove stale connections (send timeout expired)"""
        for con in list(self._connections):
            if con.is_stale():
                self._log.info(
                    "(%s:%d): send timeout", *con.get_address())
                self._remove_connection(con)

    def send(self, data):
        """Send data to connection"""
        for con in self._connections:
            con.send(data)



"""Tests for ConnectionTelnet class"""

import unittest
from unittest.mock import Mock

from ser2tcp.connection_telnet import ConnectionTelnet


class MockSocket:
    """Mock socket for testing"""
    def __init__(self):
        self.sent_data = bytearray()
        self.closed = False
        self._fileno = 5

    def send(self, data):
        self.sent_data.extend(data)
        return len(data)

    def close(self):
        self.closed = True

    def fileno(self):
        return self._fileno


class TestConnectionTelnet(unittest.TestCase):
    def _make_connection(self):
        """Helper to create ConnectionTelnet with mock socket"""
        mock_socket = MockSocket()
        addr = ('127.0.0.1', 12345)
        mock_serial = Mock()
        log = Mock()
        conn = ConnectionTelnet(
            (mock_socket, addr),
            mock_serial,
            log=log)
        # Flush initial negotiation
        conn.flush()
        mock_socket.sent_data.clear()
        return conn, mock_serial

    def test_send_escapes_iac(self):
        """IAC bytes (0xff) should be escaped to 0xff 0xff"""
        conn, _ = self._make_connection()
        conn.send(b'\xff')
        conn.flush()
        self.assertEqual(conn.socket().sent_data, b'\xff\xff')

    def test_send_escapes_multiple_iac(self):
        conn, _ = self._make_connection()
        conn.send(b'a\xff b\xff c')
        conn.flush()
        self.assertEqual(conn.socket().sent_data, b'a\xff\xff b\xff\xff c')

    def test_on_received_plain_data(self):
        """Plain data should be forwarded to serial"""
        conn, serial = self._make_connection()
        conn.on_received(b'hello')
        serial.send.assert_called_once_with(bytearray(b'hello'))

    def test_on_received_escaped_iac(self):
        """Escaped IAC (0xff 0xff) should become single 0xff"""
        conn, serial = self._make_connection()
        conn.on_received(b'\xff\xff')
        serial.send.assert_called_once_with(bytes((0xff,)))

    def test_on_received_telnet_will_command(self):
        """TELNET WILL command should not be forwarded"""
        conn, serial = self._make_connection()
        # IAC WILL 0x01 (echo)
        conn.on_received(bytes((0xff, 0xfb, 0x01)))
        serial.send.assert_not_called()

    def test_on_received_telnet_wont_command(self):
        """TELNET WONT command should not be forwarded"""
        conn, serial = self._make_connection()
        # IAC WONT 0x01
        conn.on_received(bytes((0xff, 0xfc, 0x01)))
        serial.send.assert_not_called()

    def test_on_received_telnet_do_command(self):
        """TELNET DO command should not be forwarded"""
        conn, serial = self._make_connection()
        # IAC DO 0x01
        conn.on_received(bytes((0xff, 0xfd, 0x01)))
        serial.send.assert_not_called()

    def test_on_received_telnet_dont_command(self):
        """TELNET DONT command should not be forwarded"""
        conn, serial = self._make_connection()
        # IAC DONT 0x01
        conn.on_received(bytes((0xff, 0xfe, 0x01)))
        serial.send.assert_not_called()

    def test_on_received_mixed_data_and_command(self):
        """Data mixed with TELNET commands"""
        conn, serial = self._make_connection()
        # "hello" + IAC WILL 0x01 + "world"
        conn.on_received(b'hello\xff\xfb\x01world')
        # Should receive "hello" then "world"
        calls = serial.send.call_args_list
        self.assertEqual(len(calls), 2)
        self.assertEqual(calls[0][0][0], bytearray(b'hello'))
        self.assertEqual(calls[1][0][0], bytearray(b'world'))

    def test_on_received_subnegotiation(self):
        """TELNET subnegotiation should be handled"""
        conn, serial = self._make_connection()
        # IAC SB 0x22 (some data) IAC SE
        conn.on_received(bytes((0xff, 0xfa, 0x22, 0x01, 0x02, 0xff, 0xf0)))
        serial.send.assert_not_called()

    def test_initial_negotiation_sent(self):
        """Initial TELNET negotiation should be sent on connect"""
        mock_socket = MockSocket()
        addr = ('127.0.0.1', 12345)
        mock_serial = Mock()
        log = Mock()
        conn = ConnectionTelnet((mock_socket, addr), mock_serial, log=log)
        conn.flush()
        # Should contain IAC DO 0x22 and IAC WILL 0x01
        self.assertIn(bytes((0xff, 0xfd, 0x22)), mock_socket.sent_data)
        self.assertIn(bytes((0xff, 0xfb, 0x01)), mock_socket.sent_data)


class TestTelnetConstants(unittest.TestCase):
    def test_iac_value(self):
        self.assertEqual(ConnectionTelnet.TELNET_IAC, 0xff)

    def test_command_values(self):
        self.assertEqual(ConnectionTelnet.TELNET_WILL, 0xfb)
        self.assertEqual(ConnectionTelnet.TELNET_WONT, 0xfc)
        self.assertEqual(ConnectionTelnet.TELNET_DO, 0xfd)
        self.assertEqual(ConnectionTelnet.TELNET_DONT, 0xfe)

    def test_subnegotiation_values(self):
        self.assertEqual(ConnectionTelnet.TELNET_SB, 0xfa)
        self.assertEqual(ConnectionTelnet.TELNET_SE, 0xf0)


if __name__ == "__main__":
    unittest.main()



"""Ser2tcp
Simple proxy for connecting over TCP or telnet to serial port
"""

import argparse as _argparse
import importlib.metadata as _metadata
import json as _json
import logging as _logging
import signal as _signal

import serial.tools.list_ports as _list_ports

import ser2tcp.serial_proxy as _serial_proxy
import ser2tcp.server_manager as _server_manager

try:
    _about = _metadata.metadata("ser2tcp")
    VERSION_STR = "%s %s (%s)" % (
        _about["Name"], _about["Version"], _about["Author-email"])
except _metadata.PackageNotFoundError:
    VERSION_STR = "ser2tcp (not installed)"

DESCRIPTION_STR = VERSION_STR + """
(c) 2016-2026 by pavel.revak@gmail.com
https://github.com/cortexm/ser2tcp
"""


def list_usb_devices():
    """List USB serial devices with match attributes"""
    devices = []
    for port in _list_ports.comports():
        if port.vid is not None:
            devices.append(port)
    if not devices:
        print("No USB serial devices found")
        return
    for port in devices:
        print(f"{port.device}")
        print(f"  vid: 0x{port.vid:04X}")
        print(f"  pid: 0x{port.pid:04X}")
        if port.serial_number:
            print(f"  serial_number: {port.serial_number}")
        if port.manufacturer:
            print(f"  manufacturer: {port.manufacturer}")
        if port.product:
            print(f"  product: {port.product}")
        if port.location:
            print(f"  location: {port.location}")
        print()


def main():
    """Main"""
    parser = _argparse.ArgumentParser(description=DESCRIPTION_STR)
    parser.add_argument('-V', '--version', action='version', version=VERSION_STR)
    parser.add_argument(
        '-v', '--verbose', action='count', default=0,
        help="Increase verbosity")
    parser.add_argument(
        '-u', '--usb', action='store_true',
        help="List USB serial devices and exit")
    parser.add_argument(
        '-c', '--config',
        help="configuration in JSON format")
    args = parser.parse_args()

    if args.usb:
        list_usb_devices()
        return

    if not args.config:
        parser.error("--config is required")

    _logging.basicConfig(format='%(levelname).1s: %(message)s (%(filename)s:%(lineno)s)')
    log = _logging.getLogger('ser2tcp')
    log.setLevel((30, 20, 10)[min(2, args.verbose)])

    configuration = []
    with open(args.config, "r", encoding='utf-8') as config_file:
        configuration = _json.load(config_file)

    servers_manager = _server_manager.ServersManager()
    for config in configuration:
        servers_manager.add_server(_serial_proxy.SerialProxy(config, log))

    _signal.signal(_signal.SIGTERM, servers_manager.stop)
    _signal.signal(_signal.SIGINT, servers_manager.stop)

    servers_manager.run()
    log.info("Exiting..")



"""Serial proxy - serial port management and USB device matching"""

import fnmatch as _fnmatch
import logging as _logging

import serial as _serial
import serial.tools.list_ports as _list_ports

import ser2tcp.server as _server


class SerialProxy():
    """Serial connection manager"""
    PARITY_CONFIG = {
        'NONE': _serial.PARITY_NONE,
        'EVEN': _serial.PARITY_EVEN,
        'ODD': _serial.PARITY_ODD,
        'MARK': _serial.PARITY_MARK,
        'SPACE': _serial.PARITY_SPACE,
    }
    STOPBITS_CONFIG = {
        'ONE': _serial.STOPBITS_ONE,
        'ONE_POINT_FIVE': _serial.STOPBITS_ONE_POINT_FIVE,
        'TWO': _serial.STOPBITS_TWO,
    }
    BYTESIZE_CONFIG = {
        'FIVEBITS': _serial.FIVEBITS,
        'SIXBITS': _serial.SIXBITS,
        'SEVENBITS': _serial.SEVENBITS,
        'EIGHTBITS': _serial.EIGHTBITS,
    }
    MATCH_ATTRIBUTES = ('vid', 'pid', 'serial_number', 'manufacturer',
        'product', 'location')

    def __init__(self, config, log=None):
        self._log = log if log else _logging.Logger(self.__class__.__name__)
        self._serial = None
        self._servers = []
        self._serial_config = self.fix_serial_config(config['serial'])
        baudrate = self._serial_config.get('baudrate')
        if baudrate:
            self._log.info(
                "Serial: %s %d", self._serial_config['port'], baudrate)
        else:
            self._log.info("Serial: %s", self._serial_config['port'])
        for server_config in config['servers']:
            self._servers.append(_server.Server(server_config, self, log))

    def fix_serial_config(self, config):
        """Fix serial configuration - resolve match, convert enum values"""
        if 'port' not in config:
            if 'match' not in config:
                raise ValueError("Serial config must have 'port' or 'match'")
            config['port'] = self.find_port_by_match(config.pop('match'))
        if 'parity' in config:
            for key, val in self.PARITY_CONFIG.items():
                if config['parity'] == key:
                    config['parity'] = val
        if 'stopbits' in config:
            for key, val in self.STOPBITS_CONFIG.items():
                if config['stopbits'] == key:
                    config['stopbits'] = val
        if 'bytesize' in config:
            for key, val in self.BYTESIZE_CONFIG.items():
                if config['bytesize'] == key:
                    config['bytesize'] = val
        return config

    def find_port_by_match(self, match):
        """Find serial port by matching USB device attributes"""
        if not match:
            raise ValueError("Match criteria cannot be empty")
        for key in match:
            if key not in self.MATCH_ATTRIBUTES:
                raise ValueError(f"Unknown match attribute: {key}")
        matched_ports = []
        for port_info in _list_ports.comports():
            if self._port_matches(port_info, match):
                matched_ports.append(port_info.device)
        if not matched_ports:
            raise ValueError(f"No device found matching: {match}")
        if len(matched_ports) > 1:
            raise ValueError(
                f"Multiple devices match {match}: {matched_ports}")
        return matched_ports[0]

    def _port_matches(self, port_info, match):
        """Check if port_info matches all criteria"""
        for attr, pattern in match.items():
            value = getattr(port_info, attr, None)
            if value is None:
                return False
            # Convert vid/pid to hex string for comparison
            if attr in ('vid', 'pid') and isinstance(value, int):
                value = f"0x{value:04X}"
            else:
                value = str(value)
            # Case-insensitive wildcard matching
            if not _fnmatch.fnmatch(value.upper(), str(pattern).upper()):
                return False
        return True

    def __del__(self):
        self.close()

    def connect(self):
        """Connect to serial port"""
        if not self._serial:
            try:
                self._serial = _serial.Serial(**self._serial_config)
            except (_serial.SerialException, OSError) as err:
                self._log.warning(err)
                return False
            self._log.info(
                "Serial %s connected", self._serial_config['port'])
        return True

    def has_connections(self):
        """Check if there are any active connections"""
        for server in self._servers:
            if server.has_connections():
                return True
        return False

    def disconnect(self):
        """Disconnect serial port, but if there are no active connections"""
        if self._serial and not self.has_connections():
            self._serial.close()
            self._serial = None
            self._log.info(
                "Serial %s disconnected", self._serial_config['port'])

    def close(self):
        """Close socket and all connections"""
        while self._servers:
            self._servers.pop().close()
        self.disconnect()

    def read_sockets(self):
        """Return all sockets for reading"""
        sockets = []
        for server in self._servers:
            sockets += server.read_sockets()
        if self._serial:
            sockets.append(self._serial)
        return sockets

    def write_sockets(self):
        """Return all sockets for writing (with pending data)"""
        sockets = []
        for server in self._servers:
            sockets += server.write_sockets()
        return sockets

    def send_to_connections(self, data):
        """Send data to all connections"""
        for server in self._servers:
            server.send(data)

    def process_read(self, read_sockets):
        """Process sockets with read event"""
        for server in self._servers:
            server.process_read(read_sockets)
        if self._serial and self._serial in read_sockets:
            try:
                data = self._serial.read(size=self._serial.in_waiting)
                self._log.debug("(%s): %s", self._serial_config['port'], data)
                self.send_to_connections(data)
            except (OSError, _serial.SerialException) as err:
                self._log.warning(err)
                for server in self._servers:
                    server.close_connections()
                self.disconnect()

    def process_write(self, write_sockets):
        """Process sockets with write event"""
        for server in self._servers:
            server.process_write(write_sockets)

    def process_stale(self):
        """Remove stale connections"""
        for server in self._servers:
            server.process_stale()

    def send(self, data):
        """Send data to serial port"""
        if self._serial:
            self._serial.write(data)



"""Server manager"""

import select as _select


class ServersManager():
    """Servers manager"""
    def __init__(self):
        self._servers = []
        self._running = False

    def stop(self, _signo=None, _stack_frame=None):
        """Stop the server manager loop"""
        self._running = False

    def run(self):
        """Run the server manager loop"""
        self._running = True
        while self._running:
            self.process()
        self.close()

    def add_server(self, server):
        """Add server"""
        self._servers.append(server)

    def process(self):
        """Process all servers"""
        read_list = []
        write_list = []
        for server in self._servers:
            read_list.extend(server.read_sockets())
            write_list.extend(server.write_sockets())
        ready = _select.select(read_list, write_list, [], .1)
        read_sockets, write_sockets = ready[0], ready[1]
        for server in self._servers:
            if read_sockets:
                server.process_read(read_sockets)
            if write_sockets:
                server.process_write(write_sockets)
            server.process_stale()

    def close(self):
        """Close all servers"""
        for server in self._servers:
            server.close()
